from typing import List, Iterator, Set, Iterable, Optional, Dict

from bugzoo.core.coverage import FileLine


class Snippet(object):
    """
    Represents a donor code snippet.
    """
    def __init__(self, content: str) -> None:
        self.__content = content

    @property
    def content(self) -> str:
        """
        Returns the contents of the snippet as a string.
        """
        return self.__content

    def __str__(self) -> str:
        return self.__content

    def __eq__(self, other) -> bool:
        return isinstance(other, Snippet) and self.content == other.content

    def __hash__(self) -> int:
        return hash(self.__content)


class SnippetDatabase(object):
    # TODO: implement load and save functionality (use cPickle)

    def __init__(self) -> None:
        self.__snippets = set() # type: Set[Snippet]
        self.__snippets_by_file = {} # type: Dict[str, Set[Snippet]]

    def __iter__(self) -> Iterator[Snippet]:
        """
        Returns an iterator over the snippets contained in this databse.
        """
        return self.__snippets.__iter__()

    def __len__(self) -> int:
        """
        Returns the number of unique snippets contained within the database.
        """
        return len(self.__snippets)

    def in_file(self, fn: str) -> Iterator[Snippet]:
        """
        Returns an iterator over all of the snippets that were sourced from
        a given file.
        """
        if fn in self.__snippets_by_file:
            yield from self.__snippets_by_file[fn]
        return

    def add(self,
            snippet: Snippet,
            *,
            origin: Optional[FileLine] = None
            ) -> None:
        """
        Adds a snippet to this database in-place.

        Parameters:
            snippet: the snippet to add.
            origin: an optional parameter that may be used to specify the
                origin of the snippet.

        Returns:
            nothing.
        """
        self.__snippets.add(snippet)

        if origin is not None:
            if origin.filename not in self.__snippets_by_file:
                self.__snippets_by_file[origin.filename] = {
                    Snippet('return;'),
                    Snippet('return true;'),
                    Snippet('return false;')
                }
            self.__snippets_by_file[origin.filename].add(snippet)


class SnippetFinder(object):
    def __init__(self, database: SnippetDatabase) -> None:
        self.__database = database

    def __next__(self) -> Iterator[Snippet]:
        return self.__database.__iter__()



from typing import List, Optional, Tuple, Callable
from datetime import timedelta
import logging

import bugzoo
from bugzoo.core.fileline import FileLine
from bugzoo.localization import SuspiciousnessMetric

from .problem import Problem
from .candidate import Candidate
from .searcher import Searcher
from .generator import DeletionGenerator, \
                       ReplacementGenerator, \
                       AppendGenerator, \
                       SingleEditPatches, \
                       AllTransformationsAtLine, \
                       SampleByLocalization

logger = logging.getLogger(__name__)


__all__ = ['RepairReport', 'repair']


class RepairReport(object):
    def __init__(self,
                 num_candidate_evals: int,
                 num_test_evals: int,
                 duration: timedelta) -> None:
        self.__num_candidate_evals = num_candidate_evals
        self.__num_test_evals = num_test_evals
        self.__duration = duration

    @property
    def num_candidate_evals(self) -> int:
        """
        Number of candidate evaluations performed during the search.
        """
        return self.__num_candidate_evals

    @property
    def num_test_evals(self) -> int:
        """
        Number of test-case evaluations performed during the search.
        """
        return self.__num_test_evals

    @property
    def duration(self) -> timedelta:
        """
        The wall-clock duration of the search process.
        """
        return self.__duration


def repair(bugzoo: bugzoo.BugZoo,
           problem: Problem,
           seed: Optional[int] = None,
           threads: int = 1,
           terminate_early: Optional[bool] = True,
           time_limit: Optional[timedelta] = None,
           logger: Optional[logging.Logger] = None
           ) -> Tuple[List[Candidate], RepairReport]:
    """
    Attempts to repair a given program.

    Returns:
        A tuple of the form `(patches, report)`, where `patches` is a list
        of all the candidate patches discovered during the search, and
        `report` contains a summary of the search process.
    """
    # line = FileLine("ArduCopter/Log.cpp", 577)
    # transformations = \
    #     AllTransformationsAtLine(line, problem.snippets)
    transformations = \
        SampleByLocalization(problem,
                             problem.localization,
                             problem.snippets,
                             randomize=False)
    candidates = SingleEditPatches(transformations)

    searcher = Searcher(bugzoo,
                        problem,
                        candidates,
                        threads=threads,
                        time_limit=time_limit)

    if terminate_early:
        try:
            repairs = [next(searcher)]
        except StopIteration:
            print("stopping iteration")
            repairs = []
    else:
        repairs = list(searcher)

    #for candidate in searcher.outcomes:
    #    outcome = searcher.outcomes[candidate]
    #    if not outcome.build.successful:
    #        print("FAILED TO COMPILE: {}".format(candidate))
    #    else:
    #        try:
    #            failed = next(t for t in outcome.tests if not outcome.tests[t].successful)
    #            print("FAILED TEST ({}): {}".format(failed, candidate))
    #        except StopIteration:
    #            print("SUCCESS: {}".format(candidate))

    report = RepairReport(searcher.num_candidate_evals,
                          searcher.num_test_evals,
                          searcher.time_running)
    return (repairs, report)



from typing import Iterable, Iterator, Optional, List
from mypy_extensions import NoReturn
from timeit import default_timer as timer
import logging
import datetime
import threading
import time
import signal

import bugzoo

from .candidate import Candidate
from .problem import Problem
from .outcome import OutcomeManager

logger = logging.getLogger(__name__)  # type: logging.Logger

__all__ = ['Searcher']


class Shutdown(Exception):
    pass


class Searcher(object):
    def __init__(self,
                 bugzoo: bugzoo.BugZoo,
                 problem: Problem,
                 candidates: Iterable[Candidate],
                 *,
                 threads: int = 1,
                 time_limit: Optional[datetime.timedelta] = None
                 ) -> None:
        """
        Constructs a new searcher for a given source of candidate patches.

        Parameters:
            bugzoo: a connection to the BugZoo server that should be used to
                evaluate candidate patches.
            problem: a description of the problem.
            candidates: a source of candidate patches.
            threads: the number of threads that should be made available to
                the search process.
            time_limit: an optional limit on the amount of time given to the
                searcher.
        """
        logger.debug("constructed searcher")
        assert time_limit is None or time_limit > datetime.timedelta(), \
            "if specified, time limit should be greater than zero."

        self.__bugzoo = bugzoo
        self.__problem = problem
        self.__candidates = candidates
        self.__time_limit = time_limit
        self.__num_threads = threads
        self.__outcomes = OutcomeManager()


        # records the time at which the current iteration begun
        self.__time_iteration_begun = None  # type: ignore

        self.__lock_candidates = threading.Lock()  # type: threading.Lock
        self.__counter_candidates = 0
        self.__counter_tests = 0
        self.__exhausted_candidates = False
        self.__time_running = datetime.timedelta()
        self.__error_occurred = False
        self.__found_patches = []  # type: List[Candidate]
        logger.debug("constructed searcher")

    @property
    def outcomes(self) -> OutcomeManager:
        """
        Provides a log of the outcomes of candidate patch build attempts and
        test executions.
        """
        return self.__outcomes

    @property
    def paused(self) -> bool:
        """
        Indicates whether this searcher is paused.
        """
        return (self.__found_patches != []) or self.exhausted

    @property
    def exhausted(self) -> bool:
        """
        Indicates whether or not the resources available to this searcher have
        been exhausted.
        """
        if self.__error_occurred:
            return True
        if self.__exhausted_candidates:
            return True
        if self.__time_limit is None:
            return False
        if self.time_running > self.__time_limit:
            return True
        return False

    @property
    def num_test_evals(self) -> int:
        """
        The number of test case evaluations that have been performed during
        this search process.
        """
        return self.__counter_tests

    @property
    def num_candidate_evals(self) -> int:
        """
        The number of candidate patches that have been evaluated over the
        course of this search process.
        """
        return self.__counter_candidates

    @property
    def time_limit(self) -> Optional[datetime.timedelta]:
        """
        An optional limit on the length of time that may be spent searching
        for patches.
        """
        return self.__time_limit

    @property
    def time_running(self) -> datetime.timedelta:
        """
        The amount of time that has been spent searching for patches.
        """
        time_now = timer()
        duration_secs = time_now - self.__time_iteration_begun  # type: ignore
        duration_iteration = datetime.timedelta(seconds=duration_secs)  # type: datetime.timedelta
        duration_total = self.__time_running + duration_iteration
        # DEBUG
        duration_mins = duration_iteration.seconds / 60
        logger.debug("time running: {:.2f} minutes".format(duration_mins))
        return duration_total

    def stop(self) -> None:
        # FIXME just use an event to communicate when to stop
        self.__error_occurred = True

    def __iter__(self) -> Iterator[Candidate]:
        return self

    def __next__(self) -> Candidate:
        """
        Searches for the next acceptable patch.

        Returns:
            the next patch that passes all tests.

        Raises:
            StopIteration: if the search space or available resources have
                been exhausted.
        """
        # if we have patches in the buffer, return those.
        if self.__found_patches:
            return self.__found_patches.pop()

        threads = [] # type: List[threading.Thread]

        # setup signal handlers to ensure that threads are cleanly killed
        # causes a Shutdown exception to be thrown in the search loop below
        if threading.current_thread() is threading.main_thread():
            logger.debug("attaching signal handlers")
            original_handler_sigint = signal.getsignal(signal.SIGINT)
            original_handler_sigterm = signal.getsignal(signal.SIGTERM)
            signal.signal(signal.SIGINT, lambda signum, frame: self.stop())
            signal.signal(signal.SIGTERM, lambda signum, frame: self.stop())
            logger.debug("attached signal handlers")
        else:
            logger.debug("not attaching signal handlers -- not inside main thread.")  # noqa: pycodestyle

        self.__time_iteration_begun = timer()  # type: ignore

        # TODO there's a bit of a bug: any patches that were read from the
        #   generator by the worker and were still stored in its
        #   `candidate` variable will be discarded. fixable by adding a
        #   buffer to `_try_next`.
        try:
            def worker(searcher: 'Searcher') -> None:
                while True:
                    if not searcher._try_next():
                        break

            for _ in range(self.__num_threads):
                t = threading.Thread(target=worker, args=(self,))
                threads.append(t)
                t.start()

            for t in threads:
                t.join()
        except Shutdown:
            logger.info("Reached time limit without finding repair. Terminating search.")
            self.__error_occurred = True
        except Exception as e:
            self.__error_occurred = True
        finally:
            for t in threads:
                t.join()
            if threading.current_thread() is threading.main_thread():
                logger.debug("restoring original signal handlers")
                signal.signal(signal.SIGINT, original_handler_sigint)
                signal.signal(signal.SIGTERM, original_handler_sigterm)
                logger.debug("restored original signal handlers")

        duration_iteration = timer() - self.__time_iteration_begun  # type: ignore
        self.__time_running += datetime.timedelta(seconds=duration_iteration)

        if self.__found_patches:
            return self.__found_patches.pop()
        raise StopIteration

    def _try_next(self) -> bool:
        """
        Evaluates the next candidate patch.

        Returns:
            a boolean indicating whether the calling thread should continue to
            evaluate candidate patches.
        """
        if self.paused:
            return False

        self.__lock_candidates.acquire()
        try:
            candidate = next(self.__candidates) # type: ignore
        except StopIteration:
            logger.info("All candidate patches have been exhausted.")
            self.__exhausted_candidates = True
            return False
        finally:
            self.__lock_candidates.release()

        self.__counter_candidates += 1
        bz = self.__bugzoo
        logger.debug("provisioning container")
        container = bz.containers.provision(self.__problem.bug)
        logger.debug("provisioned container")
        logger_c = logger.getChild(container.uid)
        try:
            patch = candidate.diff(self.__problem)
            diff = candidate.diff(self.__problem)
            logger_c.info("evaluating: %s\n%s\n", candidate, diff)
            logger_c.debug("applying patch")
            bz.containers.patch(container, patch)
            logger_c.debug("applied patch")

            # ensure that the patch compiles
            logger_c.debug("compiling source code")
            outcome_compilation = bz.containers.compile(container)
            logger_c.debug("compilation outcome for %s:\n%s",
                           candidate,
                           outcome_compilation.response.output)
            logger_c.debug("recording build outcome")
            self.outcomes.record_build(candidate, outcome_compilation)
            logger_c.debug("record build outcome")
            if not outcome_compilation.successful:
                logger.info("failed to compile: %s", candidate)
                return True

            # for now, execute all tests in no particular order
            # TODO perform test ordering
            logger_c.debug("executing tests")
            for test in self.__problem.tests:
                cmd = self.__problem.bug.harness.command(test)[0]
                logger_c.info("executing test: %s (%s)", test.name, candidate)
                self.__counter_tests += 1
                outcome = bz.containers.test(container, test)
                logger_c.debug("* test outcome: %s (%s) [retcode=%d]\n$ %s\n%s",
                               test.name,
                               candidate,
                               outcome.response.code,
                               cmd,
                               outcome.response.output)
                self.outcomes.record_test(candidate, test.name, outcome)
                if not outcome.passed:
                    logger_c.info("* test failed: %s (%s)", test.name, candidate)
                    return True
                logger_c.info("* test passed: %s (%s)", test.name, candidate)

            # if we've found a repair, pause the search
            self.__found_patches.append(candidate)
            logger_c.info("FOUND A REPAIR: %s", candidate)
            return True

        # TODO ensure a bool is returned when an exception occurs
        finally:
            logger_c.info("evaluated: %s", candidate)
            if container:
                del bz.containers[container.uid]



import pytest
from bugzoo.core.filechar import FileCharRange, FileChar
from darjeeling.source import SourceFile


contents = """
This is an example source code file.
It
has
a
few
lines
that we can edit.

Here's another line!
""".strip()


def test_access():
    char_range = FileCharRange(FileChar("test.txt", 11),
                               FileChar("test.txt", 17))
    f_orig = SourceFile("test.txt", contents)
    assert contents[11:18] == "example"
    assert f_orig[char_range] == "example"


def test_delete():
    char_range = FileCharRange(FileChar("test.txt", 11),
                               FileChar("test.txt", 17))
    f_orig = SourceFile("test.txt", contents)

    f_mod = f_orig.delete(char_range)
    assert str(f_mod) == contents[0:11] + contents[18:]



from typing import List, Iterator, Dict

from bugzoo.core.bug import Bug
from bugzoo.core.fileline import FileLine
from bugzoo.core.filechar import FileCharRange

from .snippet import Snippet
from .source import SourceFile


class Transformation(object):
    @property
    def diff(self) -> str:
        raise NotImplementedError


class CharRangeTransformation(Transformation):
    def __init__(self, char_range: FileCharRange) -> None:
        self.__char_range = char_range

    @property
    def char_range(self) -> FileCharRange:
        return self.__char_range


class DeleteTransformation(CharRangeTransformation):
    def __str__(self) -> str:
        return "DELETE[{}]".format(self.char_range)


class ReplaceTransformation(CharRangeTransformation):
    """
    Replaces a numbered line in a given file with a provided snippet.
    """
    def __init__(self, char_range: FileCharRange, snippet: Snippet) -> None:
        super().__init__(char_range)
        self.__snippet = snippet

    @property
    def snippet(self) -> Snippet:
        return self.__snippet

    def __str__(self) -> str:
        return "REPLACE[{}; {}]".format(self.char_range, self.__snippet)


class AppendTransformation(CharRangeTransformation):
    """
    Appends a given snippet to a specific line in a given file.
    """
    def __init__(self, char_range: FileCharRange, snippet: Snippet) -> None:
        super().__init__(char_range)
        self.__snippet = snippet

    @property
    def snippet(self) -> Snippet:
        return self.__snippet

    def __str__(self) -> str:
        return "APPEND[{}; {}]".format(self.char_range, self.__snippet)



from typing import Iterator, List, Iterable, Tuple, Optional
import random

from bugzoo.localization import Localization
from bugzoo.core.fileline import FileLine
from bugzoo.core.filechar import FileCharRange

from .problem import Problem
from .snippet import Snippet, SnippetDatabase
from .candidate import Candidate
from .transformation import Transformation, \
                            AppendTransformation, \
                            DeleteTransformation, \
                            ReplaceTransformation


class CandidateGenerator(Iterable):
    """
    Candidate generators are used to generate (normally in a lazy fashion) a
    stream of candidate patches. For now, candidate generators implement
    one-way communications: they do not accept inputs.
    """
    def __iter__(self) -> Iterator[Candidate]:
        return self

    def __next__(self) -> Candidate:
        raise NotImplementedError


class TransformationGenerator(Iterable):
    def __iter__(self) -> Iterator[Transformation]:
        return self

    def __next__(self) -> Transformation:
        raise NotImplementedError


class SingleEditPatches(CandidateGenerator):
    """
    Provides a stream of single-transformation candidate patches composed using
    a provided stream of transformations.
    """
    def __init__(self,
                 transformations: Iterator[Transformation]
                 ) -> None:
        self.__transformations = transformations

    def __next__(self) -> Candidate:
        try:
            transformation = next(self.__transformations)
        except StopIteration:
            raise StopIteration
        return Candidate([transformation])


class TargetSnippetGenerator(Iterable):
    def __init__(self,
                 targets: Iterator[FileCharRange],
                 snippets: SnippetDatabase
                 ) -> None:
        self.__targets = targets
        self.__snippets = snippets
        self.__current_target = None # type: Optional[FileCharRange]
        self.__snippets_at_target = iter([]) # type: Iterator[Snippet]

    def __iter__(self) -> Iterator[Tuple[FileCharRange, Snippet]]:
        return self

    def __next__(self) -> Tuple[FileCharRange, Snippet]:
        # fetch the next snippet at the current line
        # if there are no snippets left at this line, move onto
        # the next line. if there are no lines left, stop iterating.
        try:
            snippet = next(self.__snippets_at_target)
            return (self.__current_target, snippet)
        except StopIteration:
            try:
                # TODO use snippet generator here
                # - return snippets at current file
                # - add use/def restrictions
                self.__current_target = next(self.__targets)
                self.__snippets_at_target = \
                    self.__snippets.in_file(self.__current_target.filename)
                return self.__next__()
            except StopIteration:
                raise StopIteration


class DeletionGenerator(TransformationGenerator):
    def __init__(self, targets: Iterable[FileCharRange]) -> None:
        """
        Constructs a deletion generator.
        """
        self.__targets = reversed(list(targets))

    def __next__(self) -> Transformation:
        """
        Returns the next deletion transformation from this generator.
        """
        try:
            next_target = next(self.__targets)
        except StopIteration:
            raise StopIteration

        # TODO add static analysis
        # should we delete this line?
        # * don't delete declarations
        return DeleteTransformation(next_target)


class ReplacementGenerator(TransformationGenerator):
    """
    Uses a provided snippet database to generate all legal replacement
    transformations for a sequence of transformation targets.
    """
    def __init__(self,
                 targets: Iterator[FileCharRange],
                 snippets: SnippetDatabase
                 ) -> None:
        self.__generator_target_snippet = \
            TargetSnippetGenerator(targets, snippets)

    def __next__(self) -> Transformation:
        try:
            target, snippet = next(self.__generator_target_snippet)
        except StopIteration:
            raise StopIteration

        # TODO additional static analysis goes here
        # don't replace line with an equivalent

        return ReplaceTransformation(target, snippet)


class AppendGenerator(TransformationGenerator):
    def __init__(self,
                 targets: Iterator[FileCharRange],
                 snippets: SnippetDatabase
                 ) -> None:
        self.__generator_target_snippet = \
            TargetSnippetGenerator(targets, snippets)

    def __next__(self) -> Transformation:
        try:
            target, snippet = next(self.__generator_target_snippet)
        except StopIteration:
            raise StopIteration

        # TODO additional static analysis goes here
        # * don't append after a return
        # * don't append after a break statement

        return AppendTransformation(target, snippet)


class AllTransformationsAtLine(TransformationGenerator):
    """
    Provides a stream of all the possible transformations that can be made at
    a single line using a given snippet database.
    """
    def __init__(self,
                 problem: Problem,
                 line: FileLine,
                 snippets: SnippetDatabase,
                 *,
                 randomize: bool = True
                 ) -> None:
        # transform line to character range
        # TODO tidy this hack
        char_range = problem.sources[line.filename].line_to_char_range(line)

        # TODO clean up iterator ugliness
        self.__sources = [
            DeletionGenerator(iter([char_range])),
            ReplacementGenerator(iter([char_range]), snippets),
            AppendGenerator(iter([char_range]), snippets)
        ] # type: List[TransformationGenerator]

        # TODO implement randomize

    def __next__(self) -> Transformation:
        # TODO random/deterministic
        # choose a source at random
        try:
            source = random.choice(self.__sources)
        except IndexError:
            raise StopIteration

        # attempt to fetch a transformation from that source
        # if the source is exhausted, discard it and try again
        try:
            return next(source)
        except StopIteration:
            self.__sources.remove(source)
            return self.__next__()


# TODO: map from transformation targets to line numbers
class SampleByLocalization(TransformationGenerator):
    def __init__(self,
                 problem: Problem,
                 localization: Localization,
                 snippets: SnippetDatabase,
                 *,
                 randomize: bool = True
                 ) -> None:
        self.__localization = localization
        self.__transformations_by_line = {
            line: AllTransformationsAtLine(problem, line, snippets, randomize=randomize)
            for line in localization
        }

    def __next__(self) -> Transformation:
        try:
            line = self.__localization.sample()
            print("Looking at line: {}".format(line))
            source = self.__transformations_by_line[line]
        except ValueError:
            raise StopIteration

        try:
            return next(source)
        except StopIteration:
            del self.__transformations_by_line[line]
            self.__localization = self.__localization.without_line(line)
            return self.__next__()



from typing import Dict, Optional, Iterator

from bugzoo.testing.base import TestOutcome as BugZooTestOutcome
from bugzoo.compiler import CompilationOutcome as BugZooBuildOutcome

from .candidate import Candidate

__all__ = [
    'BuildOutcome',
    'TestOutcome',
    'TestOutcomeSet',
    'CandidateOutcome',
    'OutcomeManager'
]


class BuildOutcome(object):
    """
    Records the outcome of an attempted build.
    """
    def __init__(self, successful: bool, time_taken: float) -> None:
        self.__time_taken = time_taken
        self.__successful = successful

    @property
    def time_taken(self) -> float:
        return self.__time_taken

    @property
    def successful(self) -> bool:
        return self.__successful


class TestOutcome(object):
    """
    Records the outcome of a test execution.
    """
    def __init__(self, successful: bool, time_taken: float) -> None:
        self.__successful = successful
        self.__time_taken = time_taken

    @property
    def successful(self) -> bool:
        return self.__successful

    @property
    def time_taken(self) -> float:
        return self.__time_taken


class TestOutcomeSet(object):
    """
    Records the outcome of different test executions for a single patch.
    """
    def __init__(self,
                 outcomes: Optional[Dict[str, TestOutcome]] = None
                 ) -> None:
        if outcomes is None:
            outcomes = {}
        self.__outcomes = outcomes # type: Dict[str, TestOutcome]

    def __iter__(self) -> Iterator[str]:
        return self.__outcomes.keys().__iter__()

    def __getitem__(self, test: str) -> TestOutcome:
        return self.__outcomes[test]

    def with_outcome(self, test: str, outcome: TestOutcome) -> 'TestOutcomeSet':
        outcomes = self.__outcomes.copy()
        outcomes[test] = outcome
        return TestOutcomeSet(outcomes)


class CandidateOutcome(object):
    """
    Records the outcome of a candidate patch evaluation.
    """
    def __init__(self, build: BuildOutcome, tests: TestOutcomeSet) -> None:
        self.__build = build
        self.__tests = tests

    @property
    def build(self) -> BuildOutcome:
        return self.__build

    @property
    def tests(self) -> TestOutcomeSet:
        return self.__tests

    def with_test_outcome(self,
                          test: str,
                          successful: bool,
                          time_taken: float
                          ) -> 'CandidateOutcome':
        outcome = TestOutcome(successful, time_taken)
        test_outcomes = self.__tests.with_outcome(test, outcome)
        return CandidateOutcome(self.__build, test_outcomes)


class OutcomeManager(object):
    # FIXME hash candidate outcomes
    def __init__(self) -> None:
        self.__outcomes = {} # type: Dict[Candidate, CandidateOutcome]

    def __getitem__(self, candidate: Candidate) -> CandidateOutcome:
        return self.__outcomes[candidate]

    def __iter__(self) -> Iterator[Candidate]:
        """
        Returns an iterator over the candidate patches whose outcomes are
        stored by this manager.
        """
        return self.__outcomes.keys().__iter__()

    def record_build(self,
                     candidate: Candidate,
                     build_outcome: BugZooBuildOutcome
                     ) -> None:
        b = BuildOutcome(build_outcome.successful,
                         build_outcome.response.duration)
        c = CandidateOutcome(b, TestOutcomeSet())
        self.__outcomes[candidate] = c

    def record_test(self,
                    candidate: Candidate,
                    test_id: str,
                    test_outcome: BugZooTestOutcome
                    ) -> None:
        # TODO  race condition if there can be simultaneous test evaluations
        #       for a given patch; for now, that's not possible.
        candidate_outcome = self.__outcomes[candidate]
        candidate_outcome = candidate_outcome.with_test_outcome(test_id,
                                                                test_outcome.passed,
                                                                test_outcome.duration)
        self.__outcomes[candidate] = candidate_outcome



import logging

from .version import __version__
from .problem import Problem
import darjeeling.exceptions

logging.getLogger(__name__).setLevel(logging.INFO)
logging.getLogger(__name__).addHandler(logging.NullHandler())



from typing import List, Optional, Dict, Iterator, Callable, Set
from timeit import default_timer as timer
import tempfile
import logging
import os

import bugzoo
import bugzoo.localization.suspiciousness as metrics
from bugzoo.core.fileline import FileLine, FileLineSet
from bugzoo.core.bug import Bug
from bugzoo.core.patch import Patch
from bugzoo.core.coverage import TestSuiteCoverage
from bugzoo.core.spectra import Spectra
from bugzoo.localization import SuspiciousnessMetric, Localization
from bugzoo.testing import TestCase
from bugzoo.util import indent

import darjeeling.filters as filters
from .snippet import SnippetDatabase, Snippet
from .source import SourceFile, SourceFileCollection
from .util import get_file_contents
from .exceptions import NoFailingTests, NoImplicatedLines


class Problem(object):
    """
    Used to provide a description of a problem (i.e., a bug), and to hold
    information pertinent to its solution (e.g., coverage, transformations).
    """
    def __init__(self,
                 bz: bugzoo.BugZoo,
                 bug: Bug,
                 *,
                 suspiciousness_metric: Optional[SuspiciousnessMetric] = None,
                 in_files: List[str],
                 restrict_to_lines: Optional[FileLineSet] = None,
                 cache_coverage: bool = True,
                 verbose: bool = False,
                 logger: Optional[logging.Logger] = None,
                 line_coverage_filters: Optional[List[Callable[[str], bool]]] = None
                 ) -> None:
        """
        Constructs a Darjeeling problem description.

        Params:
            bug: A description of the faulty program.
            in_files: An optional list that can be used to restrict the set of
                transformations to those that occur in any files belonging to
                that list. If no list is provided, all source code files will
                be included.
            in_functions: An optional list that can be used to restrict the set
                of transformations to those that occur in any function whose
                name appears in the given list. If no list is provided, no
                filtering of transformations based on the function to which
                they belong will occur.

        Raises:
            NoFailingTests: if the program under repair has no failing tests.
            NoImplicatedLines: if no lines are implicated by the coverage
                information and the provided suspiciousness metric.
        """
        assert len(in_files) > 0
        self.__bug = bug
        self.__verbose = verbose

        # establish logging mechanism
        # * stream logging information to stdout
        if logger:
            self.__logger = logger
        else:
            self.__logger = \
                logging.getLogger('darjeeling.problem').getChild(bug.name)
            self.__logger.setLevel(logging.DEBUG)
            self.__logger.addHandler(logging.StreamHandler())
        self.__logger.debug("creating problem for bug: %s", bug.name)

        if suspiciousness_metric is None:
            self.__logger.debug("no suspiciousness metric provided: using Tarantula as a default.")
            suspiciousness_metric = metrics.tarantula

        # fetch coverage information
        if cache_coverage:
            self.__logger.debug("fetching coverage information from BugZoo")
            self.__coverage = bz.bugs.coverage(bug)
            self.__logger.debug("fetched coverage information from BugZoo")
        else:
            self.__logger.debug("computing coverage information")
            try:
                container = bz.containers.provision(bug)
                self.__coverage = bz.containers.coverage(container)
            finally:
                del bz.containers[container.uid]
            self.__logger.debug("computed coverage information")
        self._dump_coverage()

        # restrict coverage information to specified files
        self.__logger.debug("restricting coverage information to files:\n* %s",
                            '\n* '.join(in_files))
        self.__coverage = self.__coverage.restricted_to_files(in_files)
        self.__logger.debug("restricted coverage information.")
        self._dump_coverage()


        # determine the passing and failing tests by using coverage information
        self.__logger.debug("using test execution used to generate coverage to determine passing and failing tests")
        self.__tests_failing = set() # type: Set[TestCase]
        self.__tests_passing = set() # type: Set[TestCase]
        for test_name in self.__coverage:
            test = bug.harness[test_name]
            test_coverage = self.__coverage[test_name]
            if test_coverage.outcome.passed:
                self.__tests_passing.add(test)
            else:
                self.__tests_failing.add(test)

        self.__logger.info("determined passing and failing tests")
        self.__logger.info("* passing tests: %s", ', '.join([t.name for t in self.__tests_passing]))
        self.__logger.info("* failing tests: %s", ', '.join([t.name for t in self.__tests_failing]))
        if not self.__tests_failing:
            raise NoFailingTests

        # determine the implicated lines
        # 0. we already restricted to lines that occur in specified files
        # 1. restrict to lines covered by failing tests
        # 3. restrict to lines with suspiciousness greater than zero
        # 4. restrict to lines that are optionally provided
        self.__logger.info("Determining implicated lines")
        self.__lines = self.__coverage.failing.lines

        if restrict_to_lines is not None:
            self.__lines = self.__lines.intersection(restrict_to_lines)

        # cache contents of the implicated files
        t_start = timer()
        self.__logger.debug("storing contents of source code files")
        self.__sources = SourceFileCollection.from_bug(bz, bug, self.__lines.files)
        duration = timer() - t_start
        self.__logger.debug("stored contents of source code files (took %.1f seconds)",
                            duration)

        # restrict attention to statements
        # FIXME for now, we approximate this -- going forward, we can use
        #   Rooibos to determine transformation targets
        num_lines_before_filtering = len(self.__lines)
        self.__logger.debug("filtering lines according to content filters")
        line_content_filters = [
            filters.ends_with_semi_colon,
            filters.has_balanced_delimiters
        ]
        if line_coverage_filters:
            line_content_filters += line_coverage_filters # type: ignore
        for f in line_content_filters:
            fltr_line = \
                lambda fl: f(self.__sources.line(fl.filename, fl.num))
            self.__lines = self.__lines.filter(fltr_line)
        num_lines_after_filtering = len(self.__lines)
        num_lines_removed_by_filtering = \
            num_lines_after_filtering - num_lines_before_filtering
        self.__logger.debug("filtered lines according to content files: removed %d lines",  # noqa: pycodestyle
                            num_lines_removed_by_filtering)

        # compute fault localization
        self.__logger.info("computing fault localization")
        self.__coverage = \
            self.__coverage.restricted_to_files(self.__lines.files)
        self.__logger.debug("restricted coverage to implicated files")
        self._dump_coverage()
        self.__spectra = Spectra.from_coverage(self.__coverage)
        self.__logger.debug("generated coverage spectra: %s", self.__spectra)
        self.__localization = \
            Localization.from_spectra(self.__spectra, suspiciousness_metric)
        self.__logger.debug("transformed spectra to fault localization: %s",
                            self.__localization)
        self.__localization = \
            self.__localization.restricted_to_lines(self.__lines)
        self.__logger.info("removing non-suspicious lines from consideration")
        num_lines_before = len(self.__lines)
        self.__lines = \
            self.__lines.filter(lambda l: self.__localization.score(l) > 0)
        num_lines_after = len(self.__lines)
        num_lines_removed = num_lines_before - num_lines_after
        self.__logger.info("removed %d non-suspicious lines from consideration",
                           num_lines_removed)

        # report implicated lines and files
        self.__logger.info("implicated lines [%d]:\n%s",
                           len(self.__lines), self.__lines)
        self.__logger.info("implicated files [%d]:\n* %s",
                           len(self.__lines.files),
                           '\n* '.join(self.__lines.files))
        if len(self.__lines) == 0:
            raise NoImplicatedLines

        # construct the snippet database from the parts of the program that
        # were executed by the test suite (both passing and failing tests)
        # TODO add a snippet extractor component
        self.__logger.info("constructing snippet database")
        self.__snippets = SnippetDatabase()

        # TODO allow additional snippet filters to be provided as params
        snippet_filters = [
            filters.ends_with_semi_colon,
            filters.has_balanced_delimiters
        ]

        for line in self.__coverage.lines:
            content = self.__sources.line(line.filename, line.num).strip()
            if all(fltr(content) for fltr in snippet_filters):
                # self.__logger.debug("* found snippet at %s: %s", line, content)
                snippet = Snippet(content)
                self.__snippets.add(snippet, origin=line)

        self.__logger.info("construct snippet database: %d snippets",
                           len(self.__snippets))
        for fn in self.__lines.files:
            self.__logger.info("* %d unique snippets in %s",
                               len(list(self.__snippets.in_file(fn))), fn)

        self.__logger.debug("reducing memory footprint by discarding extraneous data")
        self.__coverage.restricted_to_files(self.__lines.files)
        self.__spectra.restricted_to_files(self.__lines.files)
        extraneous_source_fns = \
            set(self.__sources.files) - set(self.__lines.files)
        for fn in extraneous_source_fns:
            self.__sources = self.__sources.without_file(fn)
        self.__logger.debug("finished reducing memory footprint")

    def _dump_coverage(self) -> None:
        self.__logger.debug("[COVERAGE]\n%s\n[/COVERAGE]",
                            indent(str(self.__coverage), 2))

    @property
    def snippets(self) -> SnippetDatabase:
        """
        The snippet database that should be used to generate new code.
        """
        return self.__snippets

    @property
    def bug(self) -> Bug:
        """
        A description of the bug, provided by BugZoo.
        """
        return self.__bug

    @property
    def tests(self) -> Iterator[TestCase]:
        """
        Returns an iterator over the tests for this problem.
        """
        for test in self.__tests_failing:
            yield test
        for test in self.__tests_passing:
            yield test

    @property
    def tests_failing(self) -> Iterator[TestCase]:
        """
        Returns an iterator over the failing tests for this problem.
        """
        return self.__tests_failing.__iter__()

    @property
    def tests_passing(self) -> Iterator[TestCase]:
        """
        Returns an iterator over the passing tests for this problem.
        """
        return self.__tests_passing.__iter__()

    @property
    def logger(self) -> logging.Logger:
        """
        The logger that should be used to log output for this problem.
        """
        return self.__logger

    @property
    def localization(self) -> Localization:
        """
        The fault localization for this problem is used to encode the relative
        suspiciousness of source code lines.
        """
        return self.__localization

    @property
    def coverage(self) -> TestSuiteCoverage:
        """
        Line coverage information for each test within the test suite for the
        program under repair.
        """
        return self.__coverage

    @property
    def spectra(self) -> Spectra:
        """
        Provides a concise summary of the number of passing and failing tests
        that cover each implicated line.
        """
        return self.__spectra

    @property
    def lines(self) -> Iterator[FileLine]:
        """
        Returns an iterator over the lines that are implicated by the
        description of this problem.
        """
        return self.__lines.__iter__()

    @property
    def sources(self) -> SourceFileCollection:
        """
        The source code files for the program under repair.
        """
        return self.__sources

    def check_sanity(self,
                     expected_to_pass: List[TestCase],
                     expected_to_fail: List[TestCase]
                     ) -> bool:
        """

        Parameters:
            expected_to_pass: a list of test cases for the program that are
                expected to pass.
            expected_to_fail: a list o test cases for the program that are
                expected to fail.

        Returns:
            True if the outcomes of the test executions match those that are
            expected by the parameters to this method.

        Raises:
            errors.UnexpectedTestOutcomes: if the outcomes of the program
                under test, observed while computing coverage information for
                the problem, differ from those that were expected.
        """
        # determine passing and failing tests
        self.__logger.debug("sanity checking...")
        raise NotImplementedError



from typing import List


def get_file_contents(fn: str) -> str:
    try:
        with open(fn, 'r') as f:
            return f.read()
    except UnicodeDecodeError:
        pass

    # let's try to decode the using latin-1 encoding
    with open(fn, 'r', encoding='latin-1') as f:
        return f.read()


def get_lines(fn: str) -> List[str]:
    """
    Attempts to return a list of all the lines in a given source code file.
    """
    return [l.rstrip('\n') for l in get_file_contents(fn).splitlines()]



#!/usr/bin/env python
import os
from glob import glob
from setuptools import setup, find_packages


path = os.path.join(os.path.dirname(__file__), 'src/darjeeling/version.py')
with open(path, 'r') as f:
    exec(f.read())


setup(
    name='darjeeling',
    version=__version__,
    description='Distributed, language-independent, compositional search-based program repair',
    long_description='TBA',
    author='Chris Timperley',
    author_email='ctimperley@cmu.edu',
    url='https://github.com/squaresLab/Darjeeling',
    license='mit',
    python_requires='>=3.5',
    install_requires=[
        'bugzoo>=2.1.7',
        'requests',
        'flask'
    ],
    include_package_data=True,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
)



from typing import List, Iterator, Dict
import difflib
import tempfile
import logging
import os

from bugzoo import BugZoo
from bugzoo.core.bug import Bug
from bugzoo.core.patch import FilePatch
from bugzoo.core.fileline import FileLine
from bugzoo.core.filechar import FileCharRange, FileChar

from .util import get_file_contents

logger = logging.getLogger(__name__)  # type: logging.Logger


class SourceFile(object):
    def __init__(self, name: str, contents: str) -> None:
        assert isinstance(contents, str)
        assert isinstance(name, str)

        self.__name = name
        self.__contents = contents

        # FIXME this will miss the last line
        line_end_at = -1
        self.__lines = [] # type: List[FileCharRange]
        while True:
            try:
                line_start_at = line_end_at + 1
                line_end_at = contents.index("\n", line_start_at)
            except ValueError:
                break
            char_range = \
                FileCharRange(FileChar(name, line_start_at),
                              FileChar(name, line_end_at))
            self.__lines.append(char_range)

    @property
    def name(self) -> str:
        """
        The name of this source code file.
        """
        return self.__name

    def line_to_char_range(self, line: FileLine) -> FileCharRange:
        return self.__lines[line.num - 1]

#    @property
#    def lines(self) -> Iterator[FileCharRange]:
#        """
#        Returns a list of the character ranges of the lines contained
#        within this file.
#        """
#        return self.__lines.copy()

    def __str__(self) -> str:
        """
        Returns the contents of this file as a string.
        """
        return self.__contents

    def __getitem__(self, char_range: FileCharRange) -> str:
        """
        Returns the source code contained within a given range.
        """
        c_start = char_range.start.offset
        c_end = char_range.stop.offset
        return self.__contents[c_start:c_end + 1]

    def insert(self, index: FileChar, text: str) -> 'SourceFile':
        # FIXME allow insertion rather than simply appending
        offset = index.offset
        contents = "{}\n{}{}".format(self.__contents[:offset],
                                     text,
                                     self.__contents[offset:])
        return SourceFile(self.name, contents)

    def delete(self, char_range: FileCharRange) -> 'SourceFile':
        """
        Returns a variant of this file without the text contained in a given
        source range.
        """
        return self.replace(char_range, '')

    def replace(self,
                char_range: FileCharRange,
                replacement: str
                ) -> 'SourceFile':
        """
        Returns a variant of this file where the text contained in a given
        source range has been replaced with a given string.
        """
        c_start = char_range.start.offset
        c_stop = char_range.stop.offset
        contents = "{}{}\n{}".format(self.__contents[:c_start],
                                     replacement,
                                     self.__contents[c_stop+1:])
        return SourceFile(self.name, contents)

    def diff(self, other: 'SourceFile') -> str:
        diff_lines = difflib.unified_diff(str(self).splitlines(True),
                                          str(other).splitlines(True),
                                          fromfile=self.name,
                                          tofile=other.name)
        return ''.join(diff_lines)


class SourceFileCollection(object):
    @staticmethod
    def from_bug(bz: BugZoo,
                 bug: Bug,
                 filenames: List[str]
                 ) -> 'SourceFileCollection':
        logger.debug("loading source files for snapshot: %s", bug.name)
        sources = {} # type: Dict[str, SourceFile]
        ctr = bz.containers.provision(bug)
        try:
            for fn in filenames:
                # fn_ctr = os.path.join(bug.source_dir, fn)
                fn_ctr = fn
                logger.debug("loading source code file [%s] at [%s]",\
                             fn, fn_ctr)
                contents = bz.files.read(ctr, fn_ctr)
                sources[fn] = SourceFile(fn, contents)
        finally:
            del bz.containers[ctr.uid]
        collection = SourceFileCollection(sources)
        logger.debug("loaded source files for snapshot: %s", bug.name)
        return collection

    def __init__(self, contents: Dict[str, SourceFile]) -> None:
        """
        Constructs a new collection of source files.

        Parameters:
            contents: the contents of the collection, given as a dictionary of
                source files indexed by filename.
        """
        self.__contents = contents

    def __iter__(self) -> Iterator[str]:
        """
        Returns an iterator over the names of the files contained within this
        collection.
        """
        yield from self.__contents.keys()

    def __getitem__(self, fn: str) -> SourceFile:
        """
        Retrieves a file from this collection.

        Parameters:
            fn: the name of the file.

        Returns:
            the requested file.

        Raises:
            KeyError: if no file with the given name exists within this
                collection.
        """
        return self.__contents[fn]

    def delete(self, char_range: FileCharRange) -> 'SourceFileCollection':
        """
        Returns a variant of this collection where the contents of a specified
        character range within one of its files has been removed.

        Parameters:
            char_range: the range of characters that should be removed.

        Returns:
            a variant of this collection with the desired modifications.

        Raises:
            KeyError: if the file described by the character range does not
                belong to this collection.
        """
        contents_new = self.__contents.copy()
        contents_new[char_range.filename] = \
            self.__contents[char_range.filename].delete(char_range)
        return SourceFileCollection(contents_new)

    def replace(self,
                char_range: FileCharRange,
                text: str
                ) -> 'SourceFileCollection':
        """
        Returns a variant of this collection where the contents of a specified
        character range within one of its files have been replaced by a given
        text.

        Parameters:
            char_range: the range of characters that should be replaced.
            text: the text that should be used as a replacement.

        Returns:
            a variant of this collection with the desired modifications.

        Raises:
            KeyError: if the file described by the character range does not
                belong to this collection.
        """
        contents_new = self.__contents.copy()
        contents_new[char_range.filename] = \
            self.__contents[char_range.filename].replace(char_range, text)
        return SourceFileCollection(contents_new)

    def insert(self,
               index: FileChar,
               text: str
               ) -> 'SourceFileCollection':
        """
        Returns a variant of this collection where a given text has been
        inserted immediately after a specified character in one of the files
        belonging to this collection.

        Parameters:
            index: the position of the character after which the given text
                should be inserted;
            text: the text that should be inserted.

        Returns:
            a variant of this collection with the desired modifications.

        Raises:
            KeyError: if the file described by the character range does not
                belong to this collection.
        """
        contents_new = self.__contents.copy()
        contents_new[index.filename] = \
            self.__contents[index.filename].insert(index, text)
        return SourceFileCollection(contents_new)

    def diff(self, other: 'SourceFileCollection') -> str:
        raise NotImplementedError

    def line(self, fn: str, num: int) -> str:
        """
        Retrieves the contents of a given line in a source code file.
        """
        assert isinstance(num, int)

        line = FileLine(fn, num)
        f_contents = self.__contents[fn]
        char_range = f_contents.line_to_char_range(line)
        return f_contents[char_range]

    @property
    def files(self) -> List[str]:
        """
        The names of the files that are represented in this collection.
        """
        return [fn for fn in self.__contents]

    def without_file(self, fn: str) -> 'SourceFileCollection':
        """
        Produces a variant of this collection of source files that does not
        contain a file with the given name. If the named file does not exist
        within this collection, then this collection is returned.

        Parameters:
            fn: the name of the file that should not appear in the variant of
                this collection of files.

        Returns:
            a variant of this file collection that does not contain a file with
            the given name.
        """
        if fn not in self.__contents:
            return self

        contents_new = self.__contents.copy()
        return SourceFileCollection(contents_new)
